/* Copyright INMOS Limited 1988,1990,1992 */

/* CMSIDENTIFIER */
static char *CMS_Id = "PRODUCT:ITEM.VARIANT-TYPE;0(DATE)";

#include <stdio.h>
#include <errno.h>
#include <string.h>

#ifndef VMS
#include <sys/types.h>
#include <sys/stat.h>
#else
#include <perror.h>
#include <unixio.h>
#endif

#include "server.h"
#include "iserver.h"
#include "pack.h"
#include "misc.h"
#include "files.h"
#include "record.h"

static char     DataBuffer[MAX_SLICE_LENGTH + 1];
#ifndef PCTCP
extern int errno;
#endif

#if PSERVER_EXTENSION
#include "xserver.h"
#ifdef MS_DOS
#define HAVE_NO_STAT
#else
#include <fcntl.h>
#include <sys/stat.h>
#include <limits.h>
#ifdef SUN
#include <unistd.h>
#endif
#endif

/*
 *
 *      bool_t ISFILDES (fd)
 *
 *  Check if transputer fd is fildes (pserver) or stream (iserver).  Return
 *  TRUE if fildes, FALSE if stream.
 *
 */

#ifndef OPEN_MAX
#   define  OPEN_MAX    _POSIX_OPEN_MAX
#endif

#define ISFILDES(fd)        ((fd) < OPEN_MAX)
/*
 *
 *      FILE *STREAM (fd)
 *
 *  Convert fd as used on transputer to stream on host.  Special fd's 0,
 *  1 and 2 indicate standard input, output and dbgmsg, resp.  Otherwise,
 *  the file descriptor as used on transputer is equal to the host's
 *  stream, except for a possibly toggled sign bit (see above).
 *
 */

#ifndef FD_XOR
#   define  FD_XOR  0
#endif

#define  STREAM(x)   ((x)==0 ? stdin :               \
                     (x)==1 ? stdout :              \
                     (x)==2 ? stderr :              \
                              (FILE *) ((x) ^ FD_XOR))

#endif


/*
 * SpOpen
 */

void            SpOpen()
{
   FILE  *fd;
   char              *name, org, mode;
   long              fileid;
   int               size;
   struct FILE_INFO  *info;

   dbgmsg("SP.OPEN");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   name = &DataBuffer[0];
   size = (int) get_slice((unsigned char *) name);
   *(name + size) = 0;
   dbgmsg("\"%s\"", name);

   org = get_8();
   dbgmsg("org %d", org);
   mode = get_8();
   dbgmsg("mode %d", mode);

   if (strlen(name) == 0) {
      put_8(ER_ERROR);
      put_count(OutCount);
      return ;
   }

   if ((org != ORG_BINARY) && (org != ORG_TEXT)) {
      put_8(ER_ERROR);
      put_count(OutCount);
      return ;
   }

   if ((mode < 1) || (mode > 6)) {
      put_8(ER_ERROR);
      put_count(OutCount);
      return ;
   }

   fd = fopen(name, OpenModes[org-1][mode-1]);
   if (fd == NULL)
      put_8(ER_ERROR);
   else {
     fileid = RememberFile(fd, org);

     if (fileid == NO_SLOT)
        put_8(ER_NORESOURCE);
     else {
        /*
         * If we're opening for append, fseek to EOF since the ANSI spec
         * says that the initial position is `implementation specific'.
         * I always wanted to write portable code !
         */
         
        if ((mode == 3) || (mode == 6))
            fseek(fd, 0L, SEEK_END);
            
        put_8(ER_SUCCESS);
        put_32(fileid);
        
        info = &FileInfo[fileid];
        info->lastop = FIOP_NONE;

        dbgmsg("fd %ld", (int) fileid);
     }
   }
   
   put_count(OutCount);
}

/*
 * SpClose
 */

void            SpClose()
{
   long  fileid;
   FILE  *fd;
   int   forg=0;
   
   dbgmsg("SP.CLOSE");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
   
   dbgmsg("fd %ld", fileid);
   

#if PSERVER_EXTENSION
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
        close (fileid) == -1 ? put_8(ER_ERROR) : put_8(ER_SUCCESS);
        put_count(OutCount); return;
    }
#endif

   fd = FindFileDes(fileid, &forg);
   if (fd == NO_FILE)
      put_8(ER_BADID);
   else {
     if (ForgetFile(fileid) == EOF)
        put_8(ER_ERROR);
     else
        put_8(ER_SUCCESS);
   }

   put_count(OutCount);
}


/*
 * SpRead
 */

void            SpRead()
{
   char              *Data;
   int               Read; 
   long              fileid;
   int               size;
   int               forg=0;
   FILE              *fd;
   struct FILE_INFO  *info;
   bool              succeeded = FALSE;
   
   dbgmsg("SP.READ");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
   
   dbgmsg("fd %ld", fileid);
   
#ifdef SUN
   if (fileid == FD_STDIN)
      ResetTerminal();
#endif /* SUN */

   size = get_16();
   dbgmsg("request %d", size);
#if PSERVER_EXTENSION
{
    FILE    *f = STREAM (fileid);

    Data = &DataBuffer[0];
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ) {
        if ((Read = read (fileid, Data, size)) != -1)
            Data[Read] = '\0';
        succeeded = TRUE;
    }
#if 0
    else if (pserverext && isatty (fileno (f))) {
        if (fgets (Data, size+1, f))
            Read = strlen (Data);
        else
            Read = 0;
        succeeded = TRUE;
    }
#endif
    if ( succeeded ){
         if ( Read == 0 )
            put_8(ER_ERROR);
         else {
            put_8(ER_SUCCESS);
            put_slice(Read, Data);
         }
         put_count(OutCount);
         return ;
    }
}
#endif

   
   fd = FindFileDes(fileid, &forg);
   if (fd == NO_FILE){
      put_8(ER_BADID);
   }
   else {
      info = &FileInfo[fileid];
      
      if (info->lastop == FIOP_WRITE) {
         put_8(ER_NOPOSN);
      }
      else {
         Data = &DataBuffer[0];
         Read = fread(Data, 1, size, fd);
         dbgmsg("read %d", Read);
         put_8(ER_SUCCESS);
         put_slice(Read, Data);
         
         info->lastop = FIOP_READ;
      }
   }

   put_count(OutCount);
   return ;
}



/*
 * SpWrite
 */

void            SpWrite()
{
   char              *Data;
   int               Written;
   long              fileid;
   FILE              *fd;
   int               size;
   int               forg=0;
   struct FILE_INFO  *info;
   
   dbgmsg("SP.WRITE");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
#if PSERVER_EXTENSION
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
         Data = &DataBuffer[0];
         size = (int) get_slice((unsigned char *) Data);
         dbgmsg("%d chars", size);
         Written = write (fileid, Data, size);
         dbgmsg("wrote %d", Written);
         put_8(Written != 0 ?ER_SUCCESS:ER_ERROR);
         put_16((unsigned long) Written);
         put_count(OutCount);
         return;
    }
#endif
   
   dbgmsg("fd %ld", fileid);
   fd = FindFileDes(fileid, &forg);
   
   if (fd == NO_FILE) {
      put_8(ER_BADID);
      put_16(0);
   }
   else {
      info = &FileInfo[fileid];
      
      if (info->lastop == FIOP_READ) {
         put_8(ER_NOPOSN);
         put_16(0);
      }
      else {
         Data = &DataBuffer[0];
         size = (int) get_slice((unsigned char *) Data);
         dbgmsg("%d chars", size);
      
#ifdef    VMS
         /* VMS RMS generates a record for *each* item  */
         if (fwrite(Data, size, 1, fd) != 1)
            Written = 0;
         else
            Written = size;
#else
         Written = fwrite(Data, 1, size, fd);
#endif
      
         if (fd == stdout)
            fflush(stdout);
      
         dbgmsg("wrote %d", Written);
         put_8(ER_SUCCESS);
         put_16((unsigned long) Written);
         
         info->lastop = FIOP_WRITE;
      }
   }
   put_count(OutCount);
}


/*
 * SpGetBlock
 */

void            SpGetBlock()
{
   char              *Data;
   int               Read;
   long              fileid;
   int               size;
   int               forg=0;
   FILE              *fd;
   struct FILE_INFO  *info;
   bool              succeeded = FALSE;
   
   dbgmsg("SP.GETBLOCK");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
   dbgmsg("fd %ld", fileid);
   size = get_16();
   dbgmsg("request %d", size);
#if PSERVER_EXTENSION
/*
 *  PSERVER extension:  If stream is connected to terminal (isatty) use
 *                      level 2 read() to ensure proper terminal behavior
 *                      (e.g. stop reading at new-line).  If stream is
 *                      connected to terminal, it is by definition not a
 *                      binary stream, so iserver transputer runtime
 *                      would not use getblock() (but gets() instead, see
 *                      below).
 */
   {
    FILE    *f = STREAM (fileid);

    Data = &DataBuffer[0];
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ) {
        if ((Read = read (fileid, Data, size)) != -1)
            Data[Read] = '\0';
        succeeded = TRUE;
    }
#if 0
    else if (pserverext && isatty (fileno (f))) {
        if (fgets (Data, size+1, f))
            Read = strlen (Data);
        else
            Read = 0;
        succeeded = TRUE;
    }
#endif
    if ( succeeded ){
         if ( Read == 0 )
            put_8(ER_ERROR);
         else {
            put_8(ER_SUCCESS);
            put_slice(Read, Data);
         }
         put_count(OutCount);
         return ;
    }
   }
#endif

   
#ifdef SUN
   if (fileid == FD_STDIN)
      ResetTerminal();
#endif /* SUN */

   fd = FindFileDes(fileid, &forg);
   if (fd == NO_FILE)
      put_8(ER_BADID);
   else {
      info = &FileInfo[fileid];
      
      if (info->lastop == FIOP_WRITE) {
         put_8(ER_NOPOSN);
      }
      else {
         Data = &DataBuffer[0];
         Read = fread(Data, 1, size, fd);
         dbgmsg("read %d", Read);
         put_8((Read == 0) ? ER_ERROR : ER_SUCCESS);
         put_slice(Read, Data);
         
         info->lastop = FIOP_READ;
      }
   }

   put_count(OutCount);
}


/*
 * SpPutBlock
 */

void            SpPutBlock()
{
   char              *Data;
   int               Written;
   long              fileid;
   FILE              *fd;
   int               size;
   int               forg=0;
   struct FILE_INFO  *info;
   
   dbgmsg("SP.PUTBLOCK");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
#if PSERVER_EXTENSION
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
         Data = &DataBuffer[0];
         size = (int) get_slice((unsigned char *) Data);
         dbgmsg("%d chars", size);
         Written = write (fileid, Data, size);
         put_8(Written != 0 ?ER_SUCCESS:ER_ERROR);
         put_16((unsigned long) Written);
         put_count(OutCount);
         return;
    }
#endif

   fd = FindFileDes(fileid, &forg);

   dbgmsg("fd %ld", fileid);
   if (fd == NO_FILE) {
      put_8(ER_BADID);
      put_16(0);
   }
   else {
      info = &FileInfo[fileid];
      
      if (info->lastop == FIOP_READ) {
         put_8(ER_NOPOSN);
         put_16(0);
      }
      else {
         Data = &DataBuffer[0];
         size = (int) get_slice((unsigned char *) Data);
         dbgmsg("%d chars", size);

#ifdef    VMS
         /* VMS RMS generates a record for *each* item  */
         if (fwrite(Data, size, 1, fd) != 1)
            Written = 0;
         else
            Written = size;
#else
         Written = fwrite(Data, 1, size, fd);
#endif
         if (fd == stdout)
            fflush(stdout);

         dbgmsg("wrote %d", Written);
         put_8((Written == 0) ? ER_ERROR : ER_SUCCESS);
         put_16((unsigned long) Written);
         
         info->lastop = FIOP_WRITE;
      }
   }

   put_count(OutCount);
}


/*
 * SpGets
 */

void            SpGets()
{
   char              *Data;
   long              fileid;
   int               size;
   int               forg=0;
   FILE              *fd;
   struct FILE_INFO  *info;

   dbgmsg("SP.GETS");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
   dbgmsg("fd %ld", fileid);

#ifdef SUN
   if (fileid == FD_STDIN)
      ResetTerminal();
#endif /* SUN */

   size = get_16();
   dbgmsg("limit %d", size);
   
   size++;
#if PSERVER_EXTENSION
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ) {
        Data = &DataBuffer[0];
        if ((read (fileid, Data, size)) == -1){
            put_8( ER_ERROR ); put_count(OutCount); return;
        }
        size = strlen(Data);
        if (*(Data + size - 1) == '\n') {
           *(Data + size) = 0;
           --size;
        }
        dbgmsg("got %d", size);
        put_8(ER_SUCCESS);
        put_slice(size, Data);
        put_count(OutCount);
        return;
    }
#endif

   
   fd = FindFileDes(fileid, &forg);
   if (fd == NO_FILE)
      put_8(ER_BADID);
   else {
      info = &FileInfo[fileid];
      
      if (info->lastop == FIOP_WRITE) {
         put_8(ER_NOPOSN);
      }
      else {
         Data = &DataBuffer[0];
         if (fgets(Data, size, fd) == NULL) {
            put_8(ER_ERROR);
         }
         else {
            size = strlen(Data);
            if (*(Data + size - 1) == '\n') {
               *(Data + size) = 0;
               --size;
            }
            dbgmsg("got %d", size);
            put_8(ER_SUCCESS);
            put_slice(size, Data);
         }

         info->lastop = FIOP_READ;
      }
   }

   put_count(OutCount);
}


/*
 * SpPuts
 */

void            SpPuts()
{
   char              *Data;
   long              fileid;
   FILE              *fd;
   int               size;
   int               forg=0;
   int               res=0;
   struct FILE_INFO  *info;
   
   dbgmsg("SP.PUTS");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
#if PSERVER_EXTENSION
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
        Data = &DataBuffer[0];
        size = (int) get_slice((unsigned char *) Data);
        dbgmsg("%d chars", size);
        put_8((write (fileid, Data, size) == -1 ||
                write (fileid, "\n", sizeof "\n") == -1) ? ER_ERROR : ER_SUCCESS);
        put_count(OutCount); return;
    }
#endif

   fd = FindFileDes(fileid, &forg);
   
   dbgmsg("fd %ld", fileid);

   if (fd == NO_FILE)
      put_8(ER_BADID);
   else {
      info = &FileInfo[fileid];
      
      if (info->lastop == FIOP_READ) {
         put_8(ER_NOPOSN);
      }
      else {
         Data = &DataBuffer[0];
         size = (int) get_slice((unsigned char *) Data);
         dbgmsg("%d chars", size);
      
         *(Data + size) = 0;
         if (size > 0)     /* VMS RTL has bug with zero length strings */
            res = fputs(Data, fd);
            
         if (res == EOF)
            put_8(ER_ERROR);
         else {
            res = fputs("\n", fd);
            if (res == EOF)
               put_8(ER_ERROR);
            else
               put_8(ER_SUCCESS);
         }
         
         info->lastop = FIOP_WRITE;
      }
   }

   put_count(OutCount);
}


/*
 * SpFlush
 */

void            SpFlush()
{
   long  fileid;
   FILE  *fd;
   int   forg=0;
   
   dbgmsg("SP.FLUSH");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
   dbgmsg("fd %ld", fileid);
   
#if PSERVER_EXTENSION
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
        put_8(ER_SUCCESS);put_count(OutCount);return;
    }
#endif

   fd = FindFileDes(fileid, &forg);
   
   if (fd == NO_FILE)
      put_8(ER_BADID);
   else {
      if (fflush(fd) == EOF)
         put_8(ER_ERROR);
      else
         put_8(ER_SUCCESS);
   }

   put_count(OutCount);
}


/*
 * SpSeek
 */

void            SpSeek()
{
   long              Offset, Origin;
   int               orig;
   long              fileid;
   FILE              *fd;
   int               forg=0;
   int               res;
   
   dbgmsg("SP.SEEK");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();

   fd = FindFileDes(fileid, &forg);
   
   dbgmsg("fd %ld, org %d", fileid, forg);

   Offset = get_32();
   dbgmsg("offset %ld", Offset);

   Origin = get_32();
   dbgmsg("orig %ld", Origin);

   switch ((int) Origin) {
      case 1:
         orig = SEEK_SET;
         break;
      case 2:
         orig = SEEK_CUR;
         break;
      case 3:
         orig = SEEK_END;
         break;
   }


#if PSERVER_EXTENSION
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
	int res;
        if ((res = lseek (fileid, Offset, orig)) == -1 )
		put_8( ER_ERROR );
	else
		put_8( ER_SUCCESS);
        put_count(OutCount); 
	return;
    }
#endif

   if (fd == NO_FILE)
      put_8(ER_BADID);
   else {
      /* Check to see if its a record structured file */
#ifdef VMS
      if ((forg == ORG_VARIABLE) || (forg == ORG_FIXED)) {
         res = vmsseek(fileid, Offset, orig);
         put_8(res);
         put_count(OutCount);
         return;
      }
#else
      if (forg == ORG_VARIABLE) {
         if (FileInfo[fileid].format == FMT_FORMATTED)
            res = form_seq_seek(fileid, Offset, orig);
         else
            res = unform_seq_seek(fileid, Offset, orig);
            
         put_8(res);
         
         put_count(OutCount);
         return;
      }
      else if (forg == ORG_FIXED) {
         Offset *= FileInfo[fileid].mrs;
      }
#endif /* VMS */

      if (fseek(fd, Offset, orig))
         put_8(ER_ERROR);
      else {
         put_8(ER_SUCCESS);
         
         FileInfo[fileid].lastop = FIOP_NONE;
      }
   }

   put_count(OutCount);
}


/*
 * SpTell
 */

void            SpTell()
{
   long              Position;
   long              fileid;
   FILE              *fd;
   int               forg=0;
   struct FILE_INFO  *info;
   
   dbgmsg("SP.TELL");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
#if PSERVER_EXTENSION
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
        Position = tell (fileid);
        if (Position == -1L)
           put_8(ER_ERROR);
        else {
           put_8(ER_SUCCESS);
           put_32(Position);
        }
        put_count(OutCount); return;
    }
#endif

   fd = FindFileDes(fileid, &forg);
   
   dbgmsg("fd %ld", fileid);

   if (fd == NO_FILE)
      put_8(ER_BADID);
   else {
      info = &FileInfo[fileid];
      if ((info->org == ORG_VARIABLE) || (info->org == ORG_FIXED))
         Position = FileInfo[fileid].recno;
      else
         Position = ftell(fd);

      if (Position == -1L)
         put_8(ER_ERROR);
      else {
         put_8(ER_SUCCESS);
         put_32(Position);
      }
   }

   put_count(OutCount);
}


/*
 * SpEof
 */

void            SpEof()
{
   long  fileid;
   FILE  *fd;
   int   forg=0;
   
   dbgmsg("SP.EOF");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
#if PSERVER_EXTENSION
/*
 *  PACT EXTENSION: If file was opened with open2 (fildes), always return
 *                  SUCCESS - it's up to the runtime to interpret the
 *                  -1 count returned by {GET|PUT}BLOCK as meaning error.
 *
 */
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
        put_8(ER_SUCCESS);put_count(OutCount); return;
    }
#endif

   fd = FindFileDes(fileid, &forg);
   
   dbgmsg("fd %ld", fileid);

   if (fd == NO_FILE)
      put_8(ER_BADID);
   else {
      if (feof(fd))
         put_8(ER_SUCCESS);
      else
         put_8(ER_ERROR);
   }

   put_count(OutCount);
}


/*
 * SpError
 */

void            SpError()
{
   long  Errno;
   char  String[128];
   long  fileid;
   FILE  *fd;
   int   size;
   int   forg=0;
   
   dbgmsg("SP.ERROR");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
#if PSERVER_EXTENSION
/*
 *  PACT extension: if stream arg is -1, return errno and string
 */
   if(pserverext && (( fileid == -1) || ISFILDES(fileid)) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
         put_8(ER_SUCCESS);
         Errno = errno;
         dbgmsg("errno %d", Errno);
         put_32(Errno);
#ifdef SUN
         String[0] = 0;
#else
         strcpy(&String[0], strerror(errno));
#endif
         dbgmsg("dbgmsg \"%s\"", String);
         size = strlen(String);
         put_slice(size, String);
         put_count(OutCount);
         return;
   }
#endif

   fd = FindFileDes(fileid, &forg);
   
   dbgmsg("fd %ld", fileid);

   if (fd == NO_FILE) {
      put_8(ER_BADID);
   }
   else {
      if (ferror(fd)) {
         put_8(ER_SUCCESS);
         Errno = errno;
         dbgmsg("errno %d", Errno);
         put_32(Errno);
#ifdef SUN
         String[0] = 0;
#else
         strcpy(&String[0], strerror(errno));
#endif
         dbgmsg("dbgmsg \"%s\"", String);
         size = strlen(String);
         put_slice(size, String);
      }
      else
         put_8(ER_ERROR);
   }
   put_count(OutCount);
}


/*
 * SpRemove
 */

void            SpRemove()
{
   int    size;
   struct stat buf;

   dbgmsg("SP.REMOVE");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   size = (int) get_slice((unsigned char *) DataBuffer);
   DataBuffer[size] = 0;
   dbgmsg("\"%s\"", DataBuffer);
   if (DataBuffer[0] == 0)
      put_8(ER_ERROR);
   else {
      /* Check to see if the file exists */
      if (stat(DataBuffer, &buf) != 0)
         put_8(ER_NOFILE);
      else {
#ifdef SUN
         if (unlink(DataBuffer))
#else
         if (remove(DataBuffer))
#endif
            put_8(ER_ERROR);
         else
            put_8(ER_SUCCESS);
      }
   }
   put_count(OutCount);
}


/*
 * SpRename
 */

void            SpRename()
{
   char  *Oldname, *Newname;
   int   size;

   dbgmsg("SP.RENAME");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;
   Oldname = &DataBuffer[0];
   size = (int) get_slice((unsigned char *) Oldname);
   *(Oldname + size) = 0;
   dbgmsg("old \"%s\"", Oldname);

   Newname = Oldname + size + 1;
   size = (int) get_slice((unsigned char *) Newname);
   *(Newname + size) = 0;
   dbgmsg("new \"%s\"", Newname);

   if (*Oldname == 0) {
      put_8(ER_ERROR);
   }
   else {
      if (*Newname == 0) {
         put_8(ER_ERROR);
      }
      else {
         if (rename(Oldname, Newname)) {
            put_8(ER_ERROR);
         }
         else {
            put_8(ER_SUCCESS);
         }
      }
   }
   put_count(OutCount);
}


/*
 * SpIsaTTY
 */
 
void SpIsaTTY()
{
   long  fileid;
   FILE  *fd;
   int   forg=0;
   
   dbgmsg("SP.ISATTY");
   
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;
   
   fileid = get_32();
#if PSERVER_EXTENSION
    if (pserverext && ISFILDES(fileid) && ( FindFileDes( fileid, &forg) == NO_FILE ) ){
        put_8( isatty (fileid) ? ER_SUCCESS : ER_ERROR );
        put_count(OutCount);
        return;
    }
#endif

   fd = FindFileDes(fileid, &forg);
   
   dbgmsg("fd %ld", fileid);
   
   if (fd == NO_FILE)
      put_8(ER_BADID);
   else if (isatty(fileno(fd))) {
         put_8(ER_SUCCESS);
         put_8(NONZERO);
   }
   else
      put_8(ER_ERROR);
      
   put_count(OutCount);
}


/*
 * SpExists
 */
 
void SpExists()
{
   struct stat info;
   int         size;

   dbgmsg("SP.EXISTS");

   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   size = (int) get_slice((unsigned char *) DataBuffer);
   DataBuffer[size] = '\0';

   dbgmsg("\"%s\"", DataBuffer);

   if (DataBuffer[0] == '\0')
      put_8(ER_ERROR);
   else {
      put_8(ER_SUCCESS);
      
      if (stat(DataBuffer, &info) == 0)
         put_8(NONZERO);
      else
         put_8(ZERO);
   }
   
   put_count(OutCount);
}


/*
 * SpErrStat
 */
 
void SpErrStat()
{
   long  errnum;
   long  fileid;
   int   messlen;
   char  str[MAX_STRING_LENGTH];
   FILE  *fd;
   int   size;
   int   forg=0;
   extern int  sys_nerr;
   extern char *sys_errlist[];
   
   dbgmsg("SP.ERRSTAT");

   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   fileid = get_32();
   fd = FindFileDes(fileid, &forg);
   
   messlen = get_16();
   if (messlen > MAX_ITEM_SIZE)
      messlen = MAX_ITEM_SIZE;
   
   dbgmsg("fd %ld, maxlen %d", fileid, messlen);

   if (ferror(fd)) {
      errnum = errno;
      dbgmsg("errno %d", errnum);
#ifdef SUN
      if (errnum > sys_nerr)
         str[0] = '\0';
      else
         strcpy(str, sys_errlist[errnum]);
#else
      strcpy(str, strerror(errno));
#endif
      size = strlen(str);
      if (size > messlen) {
         put_8(ER_TRUNCATED);
         str[messlen] = '\0';
         size = messlen;
      }
      else
         put_8(ER_SUCCESS);

      put_32(errnum);

      dbgmsg("dbgmsg \"%s\"", &str[0]);

      put_slice(size, str);
   }
   else
      put_8(ER_ERROR);

   put_count(OutCount);
}


#if PSERVER_EXTENSION

/*
 *
 *      int file_open2 ()
 *
 *  Open a file the right way, i.e. with open(2).  Supports all oflags and
 *  allows mode to be set of new file.
 *
 *  PSERVER EXTENSION
 *
 *   Args:   VEC     name                Rets:   INT32   stream
 *          WORD    oflags
 *          WORD    mode
 *
 */

void SpOpen2()
{
    int     fd;
    char    *name;
    int     spoflags;
    int     oflags;
    int     mode, size;

   dbgmsg("SP.OPEN2");

   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

    if (! pserverext){
        put_8(ER_UNIMPLEMENTED); put_count(OutCount); return;
    }

    /* Null-terminate name, check oflags and mode.                      */
    size = (int) get_slice((unsigned char *) DataBuffer);
    DataBuffer[size] = '\0';
    name = &(DataBuffer[0]);
    spoflags = get_32();
    mode = get_32();

    /* Convert spoflags to host's oflags.                               */
    oflags = 0;
    if (spoflags & SP_O_RDONLY) oflags |= O_RDONLY;
    if (spoflags & SP_O_WRONLY) oflags |= O_WRONLY;
    if (spoflags & SP_O_RDWR  ) oflags |= O_RDWR  ;
#ifdef  O_NDELAY
    if (spoflags & SP_O_NDELAY) oflags |= O_NDELAY;
#endif
    if (spoflags & SP_O_APPEND) oflags |= O_APPEND;
#ifdef  O_SYNC
    if (spoflags & SP_O_SYNC  ) oflags |= O_SYNC  ;
#endif
    if (spoflags & SP_O_CREAT ) oflags |= O_CREAT ;
    if (spoflags & SP_O_TRUNC ) oflags |= O_TRUNC ;
    if (spoflags & SP_O_EXCL  ) oflags |= O_EXCL  ;
#ifdef  O_BINARY
    if (spoflags & SP_O_BINARY) oflags |= O_BINARY;
#endif
#ifdef  O_TEXT
    if (spoflags & SP_O_TEXT)   oflags |= O_TEXT;
#endif

    /* XXX - do something with mode? */
    /* Open file.                                                       */
    /* XXX - make sure text/bin modes work under MS-DOS */
    fd = open (name, oflags, mode);

    /* Store result:  stream.                                           */
    if ( fd == -1 ) 
        put_8(ER_ERROR); 
    else 
        put_8(ER_SUCCESS);
    put_32(fd);
    put_count(OutCount);

}


/*
 *
 *      int file_access ()
 *
 *  Check if named file can be accessed in specified manner, and return
 *  success if so, and dbgmsg if not.
 *
 *  Args:   VEC     path                Rets:   -
 *          INT32   amode
 *
 */

void SpAccess()
{
    char    *path;
    int     amode, spamode;
    int     size;

   dbgmsg("SP.ACCESS");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

    if (! pserverext){
        put_8(ER_UNIMPLEMENTED); put_count(OutCount); return;
    }

    /* Get amode and null terminate filename.                           */
    size = (int) get_slice((unsigned char *) DataBuffer);
    DataBuffer[size] = '\0';
    path = DataBuffer;
    spamode = get_32();

    /* MS-DOS supports only F_OK and W_OK, values 0 and 2, resp.  I     */
    /* guess anything may be read and executed?                         */
#ifdef MS_DOS
#   define  F_OK    0
#   define  W_OK    2

    /* XXX - if SP_R_OK or SP_X_OK, use stat() to figure out more? */
#endif

    amode = 0;
    if (spamode == SP_F_OK) amode = F_OK;
    else {
#ifdef  X_OK
        if (spamode & SP_X_OK)  amode |= X_OK;
#endif
        if (spamode & SP_W_OK)  amode |= W_OK;
#ifdef  R_OK
        if (spamode & SP_R_OK)  amode |= R_OK;
#endif
    }

    /* Return success if access allowed.                                */
    if ( access (path, amode) == -1 )
       put_8(ER_ERROR);
    else
       put_8(ER_SUCCESS);
    put_count(OutCount);
}

/*
 *
 *      int file_fcntl ()
 *
 *  Do things with stream.  See SVR3 or XPG3 docs for details.
 *
 *  Args:   BYTE    command             Rets:   INT32   value
 *          INT32   stream                      BYTE    locktype
 *          INT32   arg                         BYTE    lockwhence
 *          BYTE    locktype                    INT32   lockstart
 *          BYTE    lockwhence                  INT32   locklen
 *          INT32   lockstart                   INT32   lockpid
 *          INT32   locklen
 *          INT32   lockpid
 *
 */

void SpFcntl()
{
    int             fildes;
    int             cmd;
    int             arg;
#ifndef MS_DOS
    struct flock    lock;
#endif
    int             res, value;
    unsigned char   command, locktype, lockwhence;

   dbgmsg("SP.FCNTL");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;
    if (! pserverext){
        put_8( ER_UNIMPLEMENTED ); put_count(OutCount); return;
    }

    /* Fcntl is a UNIX command - not sure how much works on non-UNIX.   */
#ifdef SUN
    /* Get info passed into the proper variables and structures.        */
    command = get_8();
    switch (command) {
    case SP_F_DUPFD:    cmd = F_DUPFD;  break;
    case SP_F_GETFD:    cmd = F_GETFD;  break;
    case SP_F_SETFD:    cmd = F_SETFD;  break;
    case SP_F_GETFL:    cmd = F_GETFL;  break;
    case SP_F_SETFL:    cmd = F_SETFL;  break;
    case SP_F_GETLK:    cmd = F_GETLK;  break;
    case SP_F_SETLK:    cmd = F_SETLK;  break;
    case SP_F_SETLKW:   cmd = F_SETLKW; break;
    default:            dbgmsg ("file_fcntl(): illegal command %d\n", command);
                        put_8( ER_ERROR);put_count(OutCount);return;
    }
    fildes = get_32();
    arg = get_32();
    locktype - get_8(); lockwhence = get_8();
    if (! ISFILDES(fildes))             /* too much hassle to support streams as well */
        {put_8( ER_ERROR);put_count(OutCount);return;}
    switch (cmd) {
    case F_DUPFD: case F_SETFD: case F_SETFL:
        break;
    case F_GETLK: case F_SETLK: case F_SETLKW:
        switch (locktype) {
        case SP_F_RDLCK:    lock.l_type = F_RDLCK;      break;
        case SP_F_WRLCK:    lock.l_type = F_WRLCK;      break;
        case SP_F_UNLCK:    lock.l_type = F_UNLCK;      break;
        default:            dbgmsg ("file_fcntl(): illegal lock type %d\n", locktype);
                            put_8( ER_ERROR);put_count(OutCount);return;
        }
        switch (lockwhence) {
        case SP_SEEK_SET:   lock.l_whence = SEEK_SET;   break;
        case SP_SEEK_CUR:   lock.l_whence = SEEK_CUR;   break;
        case SP_SEEK_END:   lock.l_whence = SEEK_END;   break;
        default:            dbgmsg ("file_fcntl(): illegal whence %d", lockwhence);
                            put_8( ER_ERROR);put_count(OutCount);return;
        }
        lock.l_start = get_32();
        lock.l_len   = get_32();
        lock.l_pid   = get_32();
        arg = (int) &lock;
        break;
    }

    /* Do fcntl, return result, or ER_ERROR if failed.                  */
    value = fcntl (fildes, cmd, arg);
    if (value == -1){
	put_8( ER_ERROR ); put_count( OutCount ); return;
    }

    /* Return lock structure.                                           */
    switch (cmd) {
    case F_GETLK: case F_SETLK: case F_SETLKW:
        switch (lock.l_type) {
        case F_RDLCK:   locktype = SP_F_RDLCK;   break;
         case F_WRLCK:   locktype = SP_F_WRLCK;   break;
        case F_UNLCK:   locktype = SP_F_UNLCK;   break;
        default:        dbgmsg ("file_fcntl(): fcntl() returned illegal lock type %d\n", lock.l_type);
                        put_8( ER_ERROR ); put_count( OutCount ); return;
        }
        switch (lock.l_whence) {
        case SEEK_SET:  lockwhence = SP_SEEK_SET;  break;
        case SEEK_CUR:  lockwhence = SP_SEEK_CUR;  break;
        case SEEK_END:  lockwhence = SP_SEEK_END;  break;
        default:        dbgmsg ("file_fcntl(): fcntl() returned illegal whence %d", lock.l_whence);
                        put_8( ER_ERROR ); put_count( OutCount ); return;
        }
        break;
    default:
        locktype = 0;
        lockwhence = 0;
        lock.l_start = lock.l_len = lock.l_pid = 0;
    }
    put_8( ER_SUCCESS );
    put_32( value );
    put_8( locktype );
    put_8( lockwhence );
    put_32( lock.l_start );
    put_32( lock.l_len );
    put_32( lock.l_pid );

    /* Return success if OK.                                            */
    put_count( OutCount );
    return;

    /* Not implemented on non-UNIX.                                     */
#else
    put_8( ER_UNIMPLEMENTED);put_count(OutCount);return;
#endif
}


/*
 *
 *      int file_stat ()
 *
 *  Return stat info of named file or opened stream (if not -1).
 *
 *  Args:   VEC     path                Rets:   INT32   mode
 *          INT32   stream                      INT32   ino
 *                                              INT32   dev
 *                                              INT32   rdev
 *                                              INT32   nlink
 *                                              INT32   uid
 *                                              INT32   gid
 *                                              INT32   size
 *                                              INT32   atime
 *                                              INT32   mtime
 *                                              INT32   ctime
 *
 */

void SpStat()
{

#ifndef HAVE_NO_STAT
    struct stat s;
#endif
    int         mode;
    int         size;
    char        *path;
    int         fd;

   dbgmsg("SP.STAT");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

    if (! pserverext){
        put_8(ER_UNIMPLEMENTED); put_count(OutCount); return;
    }

    size = (int) get_slice((unsigned char *) DataBuffer);
    DataBuffer[size] = '\0';
    path = DataBuffer;
    fd = get_32();
#ifndef HAVE_NO_STAT
     /* Fstat() if stream not -1.                                        */
    if (fd != -1) {
        if (fstat (ISFILDES(fd) ? fd : fileno(STREAM(fd)), &s) == -1)
            {put_8( ER_ERROR);put_count(OutCount);return;}
    }

    /* Or fstat() if no path.                                           */
    else {
        if (stat (path, &s) == -1)
            {put_8( ER_ERROR);put_count(OutCount);return;}
    }

    /* Convert mode from host bits to SP.                               */
    mode = 0;
    switch (s.st_mode & S_IFMT) {
    case S_IFIFO:   mode = SP_S_IFIFO;  break;
    case S_IFCHR:   mode = SP_S_IFCHR;  break;
    case S_IFDIR:   mode = SP_S_IFDIR;  break;
    case S_IFBLK:   mode = SP_S_IFBLK;  break;
    case S_IFREG:   mode = SP_S_IFREG;  break;
    default:        dbgmsg ("file_stat(): [f]stat() returned illegal mode %x\n", s.st_mode);
                    put_8( ER_ERROR);put_count(OutCount);return;
    }
    if (s.st_mode & S_ISUID)    mode |= SP_S_ISUID;
    if (s.st_mode & S_ISGID)    mode |= SP_S_ISGID;
    if (s.st_mode & S_ISVTX)    mode |= SP_S_ISVTX;
#ifdef S_ENFMT
    if (s.st_mode & S_ENFMT)    mode |= SP_S_ENFMT;
#endif
    if (s.st_mode & S_IRUSR)    mode |= SP_S_IRUSR;
    if (s.st_mode & S_IWUSR)    mode |= SP_S_IWUSR;
    if (s.st_mode & S_IXUSR)    mode |= SP_S_IXUSR;
    if (s.st_mode & S_IRGRP)    mode |= SP_S_IRGRP;
    if (s.st_mode & S_IWGRP)    mode |= SP_S_IWGRP;
    if (s.st_mode & S_IXGRP)    mode |= SP_S_IXGRP;
    if (s.st_mode & S_IROTH)    mode |= SP_S_IROTH;
    if (s.st_mode & S_IWOTH)    mode |= SP_S_IWOTH;
    if (s.st_mode & S_IXOTH)    mode |= SP_S_IXOTH;

    /* Return all fields.                                               */
    put_8( ER_SUCCESS );
    put_32(mode);
    put_32(s.st_ino);
    put_32(s.st_dev);
    put_32(s.st_rdev);
    put_32(s.st_nlink);
    put_32(s.st_uid);
    put_32(s.st_gid);
    put_32(s.st_size);
    put_32(s.st_atime);
    put_32(s.st_mtime);
    put_32(s.st_ctime);

    /* XXX - may need to fix time fields for non-DOS/UNIX */

    /* Return success.                                                  */
    put_count(OutCount);
    return;
#else

    /* Not implemented for all systems yet.                             */
    put_8( ER_UNIMPLEMENTED);put_count(OutCount);return;
#endif
}


/*
 *
 *      int file_link ()
 *
 *  Link old file to new one and return success/dbgmsg.  Always returns
 *  dbgmsg on MS-DOS, where this doesn't work.
 *
 *  Args:   VEC     old                 Rets:   -
 *          VEC     new
 *
 */

void SpLink()
{
    char    *old, *new;
    char     DataBuffer1[MAX_SLICE_LENGTH + 1];
    int      size;

   dbgmsg("SP.LINK");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

    if (! pserverext){
        put_8(ER_UNIMPLEMENTED); put_count(OutCount); return;
    }
    /* Get new and old names, and null-terminate.  Must be careful,     */
    /* because the string terminator for the old name overwrites the    */
    /* length of the new name.  So do new name first, then old name.    */
    size = (int) get_slice((unsigned char *) DataBuffer1);
    DataBuffer1[size] = '\0';
    old = DataBuffer1;
    size = (int) get_slice((unsigned char *) DataBuffer);
    DataBuffer[size] = '\0';
    new = DataBuffer;

    /* On UNIX, do link() call and return status.                       */
#ifdef SUN
    if (link (old, new) == -1 )
	 {put_8( ER_ERROR );put_count(OutCount);return;}
    else
         {put_8(  ER_SUCCESS);put_count(OutCount);return;}

    /* No such thing on MS-DOS.                                         */
#else
    put_8( ER_ERROR);put_count(OutCount);return;
#endif
}
/*
 *
 *      int file_unlink ()
 *
 *  Unlink file or directory.  Returns success if succeeded, dbgmsg otherwise.
 *
 *  Args:   VEC     path                Rets:   -
 *
 */

void SpUnlink()
{
    char    *path;
    int     size;

   dbgmsg("SP.UNLINK");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;
    if (! pserverext){
        put_8(ER_UNIMPLEMENTED); put_count(OutCount); return;
    }

    size = (int) get_slice((unsigned char *) DataBuffer);
    /* Null terminate filename.                                         */
    DataBuffer[size] = '\0';
    path = DataBuffer;

    /* On MS-DOS, call rmdir() if path is directory.                    */
#ifdef MSDOS
#if !defined(HAVE_NO_STAT)
    {
        struct stat s;
        if (stat (path, &s) == -1) {
            put_8( ER_ERROR);put_count(OutCount);return;
        if (s.st_mode & S_IFDIR){
            if ( rmdir (path) == -1 )
               put_8( ER_ERROR );
            else 
               put_8( ER_SUCCESS );
            put_count(OutCount); return;
        }
     }
#endif
#endif

    /* Unlink and return status.                                        */
    if ( unlink (path) == -1 )
       put_8( ER_ERROR );
    else
       put_8( ER_SUCCESS );
    put_count(OutCount);
    return;
}
#endif

